package database

import (
	"database/sql"
	"errors"
	"strconv"

	"strings"

	_ "github.com/go-sql-driver/mysql"
	"github.com/pivotal-cf-experimental/mysql-metrics/metrics"
)

type DatabaseClient struct {
	connection *sql.DB
}

func NewDatabaseClient(connection *sql.DB) *DatabaseClient {
	return &DatabaseClient{
		connection: connection,
	}
}

func (dc *DatabaseClient) QueryMetrics(mappings metrics.MetricMappingConfig, emitLeaderFollowerMetrics bool) ([]*metrics.Metric, error) {
	rows, err := dc.connection.Query("SHOW GLOBAL STATUS")
	if err != nil {
		return []*metrics.Metric{dc.availableMetric(false, mappings.GetInternalMetricMappings())}, err
	}
	var gatheredMetrics []*metrics.Metric
	gatheredMetrics = append(gatheredMetrics, dc.addMetrics(rows, mappings.GetMysqlMetricMappings())...)
	rows.Close()

	rows, err = dc.connection.Query("SHOW GLOBAL VARIABLES")
	if err != nil {
		return []*metrics.Metric{dc.availableMetric(false, mappings.GetInternalMetricMappings())}, err
	}
	gatheredMetrics = append(gatheredMetrics, dc.addMetrics(rows, mappings.GetMysqlMetricMappings())...)
	rows.Close()

	if emitLeaderFollowerMetrics {
		rows, err = dc.connection.Query("SHOW SLAVE STATUS")
		if err != nil {
			gatheredMetrics = append(gatheredMetrics, dc.availableMetric(false, mappings.GetInternalMetricMappings()))
			return gatheredMetrics, err
		}

		followerAvailableMetric := dc.checkFollowerAvailable(rows, mappings.GetLeaderFollowerMetricMappings())
		gatheredMetrics = append(gatheredMetrics, followerAvailableMetric)
		rows.Close()

		if followerAvailableMetric.Value == 1.0 {
			rows, err := dc.connection.Query("SHOW SLAVE STATUS")
			if err != nil {
				gatheredMetrics = append(gatheredMetrics, dc.availableMetric(false, mappings.GetInternalMetricMappings()))
				return gatheredMetrics, err
			}

			slaveStatusMetrics, err := dc.getLeaderFollowerMetrics(rows, mappings.GetLeaderFollowerMetricMappings())

			if err != nil {
				rows.Close()
				gatheredMetrics = append(gatheredMetrics, dc.availableMetric(false, mappings.GetInternalMetricMappings()))
				return gatheredMetrics, err
			}
			rows.Close()
			gatheredMetrics = append(gatheredMetrics, slaveStatusMetrics...)
		}
	}

	gatheredMetrics = append(gatheredMetrics, dc.availableMetric(true, mappings.GetInternalMetricMappings()))

	return gatheredMetrics, nil
}

func (dc *DatabaseClient) getLeaderFollowerMetrics(rows *sql.Rows, leaderFollowerMetricMappings map[string]metrics.LeaderFollowerMetricMapping) ([]*metrics.Metric, error) {
	var gatheredMetrics []*metrics.Metric

	slaveStatus, err := dc.getSlaveStatus(rows)
	if err != nil {
		return gatheredMetrics, err
	}

	for metricName, mapping := range leaderFollowerMetricMappings {
		value, found := slaveStatus[metricName]
		if found {
			metric := metrics.Metric{
				Key:      mapping.Key,
				Unit:     mapping.Unit,
				RawValue: value,
			}
			metric = dc.metricWithValue(&metric)

			gatheredMetrics = append(gatheredMetrics, &metric)
		}
	}
	return gatheredMetrics, nil
}

func (dc *DatabaseClient) getSlaveStatus(rows *sql.Rows) (map[string]string, error) {
	result := make(map[string]string)

	columns, err := rows.Columns()
	if err != nil {
		return nil, err
	}

	if rows.Next() {
		columnValues := make([]interface{}, len(columns))
		columnPointers := make([]interface{}, len(columns))
		for i := range columns {
			columnPointers[i] = &columnValues[i]
		}
		err := rows.Scan(columnPointers...)
		if err != nil {
			panic(err)
		}
		for index, col := range columns {
			value := columnValues[index]
			if value == nil {
				value = []uint8("NULL")
			}
			result[strings.ToLower(col)] = string(value.([]uint8))
		}
	}

	return result, nil
}

func (dc *DatabaseClient) checkFollowerAvailable(rows *sql.Rows, leaderFollowerMetricMappings map[string]metrics.LeaderFollowerMetricMapping) *metrics.Metric {
	isFollower := 0.0
	if rows.Next() {
		isFollower = 1.0
	}
	key := leaderFollowerMetricMappings["is_follower"].Key
	return &metrics.Metric{Key: key, Value: isFollower, Unit: "bool"}
}

func (dc *DatabaseClient) addMetrics(rows *sql.Rows, metricMappings map[string]metrics.MysqlMetricMapping) []*metrics.Metric {
	var gatheredMetrics []*metrics.Metric

	for rows.Next() {
		var key, value string

		rows.Scan(&key, &value)
		spec, found := metricMappings[strings.ToLower(key)]
		if found {
			metric := metrics.Metric{
				Key:      spec.Key,
				Unit:     spec.Unit,
				RawValue: value,
			}
			metric = dc.metricWithValue(&metric)
			gatheredMetrics = append(gatheredMetrics, &metric)
		}
	}
	return gatheredMetrics
}

func (dc *DatabaseClient) metricWithValue(metric *metrics.Metric) metrics.Metric {
	floatValue, err := dc.parseFloat(metric.RawValue)
	if err != nil {
		floatValue, err = dc.parseOnOff(metric.RawValue)
	}
	if err != nil {
		floatValue, err = dc.parseClusterStatus(metric.RawValue)
	}

	return metrics.Metric{
		Key:      metric.Key,
		Unit:     metric.Unit,
		RawValue: metric.RawValue,
		Value:    floatValue,
		HasError: err != nil,
	}
}

func (dc *DatabaseClient) availableMetric(isAvailable bool, internalMetricMappings map[string]metrics.InternalMetricMapping) *metrics.Metric {
	availableAsFloat := 0.0
	if isAvailable {
		availableAsFloat = 1.0
	}
	key := internalMetricMappings["available"].Key

	return &metrics.Metric{Key: key, Value: availableAsFloat, Unit: "bool"}
}

func (dc *DatabaseClient) parseFloat(value string) (float64, error) {
	return strconv.ParseFloat(value, 64)
}

func (dc *DatabaseClient) parseOnOff(value string) (float64, error) {
	var floatValue float64
	var err error

	switch strings.ToLower(value) {
	case "on":
		floatValue = 1.0
	case "off":
		floatValue = 0.0
	case "null":
		floatValue = -1.0
	default:
		err = errors.New("could not convert wsrep_ready")
	}

	return floatValue, err
}

func (dc *DatabaseClient) parseClusterStatus(value string) (float64, error) {
	var floatValue float64
	var err error

	switch strings.ToLower(value) {
	case "primary":
		floatValue = 1
	case "non-primary":
		floatValue = 0
	case "disconnected":
		floatValue = -1
	default:
		err = errors.New("could not convert wsrep_cluster_status")
	}

	return floatValue, err
}
