package database_test

import (
	"database/sql"
	"errors"
	"fmt"

	"github.com/DATA-DOG/go-sqlmock"
	"github.com/pivotal-cf-experimental/mysql-metrics/database"
	"github.com/pivotal-cf-experimental/mysql-metrics/metrics"

	"strings"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
)

var _ = Describe("DatabaseClient", func() {

	var (
		conn                         *sql.DB
		mock                         sqlmock.Sqlmock
		err                          error
		client                       *database.DatabaseClient
		mysqlMetricMappings          map[string]metrics.MysqlMetricMapping
		internalMetricMappings       map[string]metrics.InternalMetricMapping
		leaderFollowerMetricMappings map[string]metrics.LeaderFollowerMetricMapping
		metricMappingConfig          metrics.MetricMappingConfig
		emitLeaderFollowerMetrics    bool
	)

	findMetricByKey := func(metrics []*metrics.Metric, key string) *metrics.Metric {
		for i := range metrics {
			aMetric := metrics[i]
			if strings.Contains(aMetric.Key, key) {
				return aMetric
			}
		}
		return nil
	}

	seedEmptyGlobalVariablesResponse := func() {
		mock.ExpectQuery(`SHOW GLOBAL VARIABLES`).WillReturnRows(sqlmock.NewRows([]string{"variable_name", "value"}))
	}

	seedSlaveStatusTwice := func(rowValue string) {
		row := sqlmock.NewRows([]string{"Seconds_Behind_Master", "Slave_SQL_Running_State", "SQL_Lib_Not_Able_To_Parse"}).AddRow([]uint8(rowValue), []uint8("Slave has read all relay log; waiting for more updates"), []uint8(nil))
		mock.ExpectQuery("SHOW SLAVE STATUS").WillReturnRows(row)
		row2 := sqlmock.NewRows([]string{"Seconds_Behind_Master", "Slave_SQL_Running_State", "SQL_Lib_Not_Able_To_Parse"}).AddRow([]uint8(rowValue), []uint8("Slave has read all relay log; waiting for more updates"), []uint8(nil))
		mock.ExpectQuery("SHOW SLAVE STATUS").WillReturnRows(row2)
	}

	seedOneRowFromDatabase := func(keyName string, value string) {
		row := sqlmock.NewRows([]string{"variable_name", "value"}).AddRow(keyName, value)
		mock.ExpectQuery(`SHOW GLOBAL STATUS`).WillReturnRows(row)
		seedEmptyGlobalVariablesResponse()
	}

	BeforeEach(func() {
		mysqlMetricMappings = map[string]metrics.MysqlMetricMapping{
			"questions":                     {Key: "/p.mysql/global_statuses/questions", Unit: "metric"},
			"innodb_buffer_pool_pages_free": {Key: "/p.mysql/innodb/buffer_pool_pages_free", Unit: "byte"},
			"innodb_data_read":              {Key: "/p.mysql/innodb/data_read", Unit: "byte"},
			"wsrep_cluster_status":          {Key: "/p.mysql/galera/wsrep_cluster_status", Unit: "number"},
			"wsrep_ready":                   {Key: "/p.mysql/galera/wsrep_ready", Unit: "bool"},
			"max_connections":               {Key: "/p.mysql/variables/max_connections", Unit: "integer"},
		}
		internalMetricMappings = map[string]metrics.InternalMetricMapping{
			"available": {Key: "/p.mysql/available"},
		}
		leaderFollowerMetricMappings = map[string]metrics.LeaderFollowerMetricMapping{
			"is_follower":           {Key: "/p.mysql/follower/is_follower", Unit: "bool"},
			"seconds_behind_master": {Key: "/p.mysql/follower/seconds_behind_master", Unit: "number"},
		}

		metricMappingConfig = metrics.MetricMappingConfig{
			MysqlMetricMappings:          mysqlMetricMappings,
			InternalMetricMappings:       internalMetricMappings,
			LeaderFollowerMetricMappings: leaderFollowerMetricMappings,
		}

		conn, mock, err = sqlmock.New()
		Expect(err).NotTo(HaveOccurred())

		client = database.NewDatabaseClient(conn)
	})

	Describe("QueryMetrics()", func() {
		Describe("when the query errors", func() {
			BeforeEach(func() {
				emitLeaderFollowerMetrics = false
				mock.ExpectQuery(`SHOW GLOBAL STATUS`).WillReturnError(errors.New("query error"))
			})

			It("returns an error and database is down metric", func() {
				resultMetrics, err := client.QueryMetrics(metricMappingConfig, emitLeaderFollowerMetrics)
				Expect(err).To(HaveOccurred())
				Expect(err).To(MatchError("query error"))
				Expect(resultMetrics).NotTo(BeNil())
				Expect(len(resultMetrics)).To(Equal(1))

				metric := resultMetrics[0]
				Expect(metric.Key).To(Equal("/p.mysql/available"))
				Expect(metric.Value).To(Equal(0.0))
				Expect(metric.HasError).To(BeFalse())
			})
		})

		Describe("when the query returns non-float data", func() {
			BeforeEach(func() {
				emitLeaderFollowerMetrics = false
				row := sqlmock.NewRows([]string{"variable_name", "value"}).
					AddRow("Questions", "Nope").
					AddRow("Innodb_buffer_pool_pages_free", "0")
				mock.ExpectQuery(`SHOW GLOBAL STATUS`).WillReturnRows(row)
				seedEmptyGlobalVariablesResponse()
			})

			It("returns the error (and the good metrics)", func() {
				resultMetrics, err := client.QueryMetrics(metricMappingConfig, emitLeaderFollowerMetrics)
				Expect(err).NotTo(HaveOccurred())
				Expect(resultMetrics).NotTo(BeNil())
				Expect(len(resultMetrics)).To(Equal(3))

				metric := findMetricByKey(resultMetrics, "global_statuses")
				Expect(metric.Key).To(Equal("/p.mysql/global_statuses/questions"))
				Expect(metric.RawValue).To(Equal("Nope"))
				Expect(metric.HasError).To(BeTrue())

				metric = findMetricByKey(resultMetrics, "buffer_pool_pages_free")
				Expect(metric.Key).To(Equal("/p.mysql/innodb/buffer_pool_pages_free"))
				Expect(metric.HasError).To(BeFalse())
			})
		})

		Describe("when the query returns expected data", func() {
			var (
				numQuestions      float64
				dataRead          float64
				maxConnections    float64
				clusterStatus     string
				galeraReadyStatus string
			)
			BeforeEach(func() {
				numQuestions = float64(125146.0)
				dataRead = float64(24121.0)
				clusterStatus = "Primary" // this might also be PRIMARY sometimes?
				galeraReadyStatus = "ON"
				maxConnections = 12
				emitLeaderFollowerMetrics = false
			})

			seedDatabase := func() {
				row := sqlmock.NewRows([]string{"variable_name", "value"}).
					AddRow("something", "doesntmatter").
					AddRow("Questions", fmt.Sprintf("%f", numQuestions)).
					AddRow("Innodb_data_read", fmt.Sprintf("%f", dataRead)).
					AddRow("wsrep_cluster_status", clusterStatus).
					AddRow("wsrep_ready", galeraReadyStatus).
					AddRow("Other Thing", "Still doesn't matter")
				mock.ExpectQuery(`SHOW GLOBAL STATUS`).WillReturnRows(row)

				row = sqlmock.NewRows([]string{"variable_name", "value"}).AddRow("max_connections", fmt.Sprintf("%f", maxConnections))
				mock.ExpectQuery(`SHOW GLOBAL VARIABLES`).WillReturnRows(row)
			}

			It("returns an array of metric objects", func() {
				seedDatabase()
				resultMetrics, err := client.QueryMetrics(metricMappingConfig, emitLeaderFollowerMetrics)

				Expect(err).NotTo(HaveOccurred())
				Expect(resultMetrics).NotTo(BeNil())
				Expect(len(resultMetrics)).To(Equal(6))

				aResultMetric := findMetricByKey(resultMetrics, "global_statuses")
				Expect(aResultMetric.Key).To(Equal("/p.mysql/global_statuses/questions"))
				Expect(aResultMetric.Value).To(Equal(numQuestions))
				Expect(aResultMetric.Unit).To(Equal("metric"))

				aResultMetric = findMetricByKey(resultMetrics, "data_read")
				Expect(aResultMetric.Key).To(Equal("/p.mysql/innodb/data_read"))
				Expect(aResultMetric.Value).To(Equal(dataRead))
				Expect(aResultMetric.Unit).To(Equal("byte"))

				aResultMetric = findMetricByKey(resultMetrics, "max_connections")
				Expect(aResultMetric.Key).To(Equal("/p.mysql/variables/max_connections"))
				Expect(aResultMetric.Value).To(Equal(maxConnections))
				Expect(aResultMetric.Unit).To(Equal("integer"))

				aResultMetric = findMetricByKey(resultMetrics, "/p.mysql/available")
				Expect(aResultMetric.Key).To(Equal("/p.mysql/available"))
				Expect(aResultMetric.Value).To(Equal(1.0))
				Expect(aResultMetric.Unit).To(Equal("bool"))

			})

			Describe("when emitLeaderFollowerMetrics is true", func() {
				Describe("when SHOW SLAVE STATUS returns a query error", func() {
					BeforeEach(func() {
						emitLeaderFollowerMetrics = true
						seedDatabase()
						mock.ExpectQuery(`SHOW SLAVE STATUS`).WillReturnError(errors.New("query error"))
					})

					It("emits the available metric as false, does not emit any leader/follower metrics, and returns an error", func() {
						resultMetrics, err := client.QueryMetrics(metricMappingConfig, emitLeaderFollowerMetrics)
						Expect(err).To(HaveOccurred())
						Expect(err).To(MatchError("query error"))
						Expect(resultMetrics).NotTo(BeNil())

						metric := findMetricByKey(resultMetrics, "/p.mysql/follower/seconds_behind_master")
						Expect(metric).To(BeNil())

						metric = findMetricByKey(resultMetrics, "/p.mysql/available")
						Expect(metric).NotTo(BeNil())
						Expect(metric.Value).To(Equal(0.0))
						Expect(metric.HasError).To(BeFalse())
					})

					Describe("if SHOW GLOBAL STATUS produces metrics", func() {
						BeforeEach(func() {
							seedDatabase()
							mock.ExpectQuery(`SHOW SLAVE STATUS`).WillReturnError(errors.New("query error"))
						})

						It("still emits those mysql metrics", func() {
							resultMetrics, err := client.QueryMetrics(metricMappingConfig, emitLeaderFollowerMetrics)
							Expect(err).To(HaveOccurred())
							Expect(err).To(MatchError("query error"))
							Expect(resultMetrics).NotTo(BeNil())

							metric := findMetricByKey(resultMetrics, "/p.mysql/variables/max_connections")
							Expect(metric).NotTo(BeNil())
							Expect(metric.Value).To(Equal(maxConnections))
							Expect(metric.Unit).To(Equal("integer"))
						})
					})
				})

				Describe("when SHOW SLAVE STATUS returns an empty result (leader node)", func() {
					BeforeEach(func() {
						emitLeaderFollowerMetrics = true
						seedDatabase()
						mock.ExpectQuery("SHOW SLAVE STATUS").WillReturnRows(sqlmock.NewRows([]string{}))
					})

					It("returns is_follower metric as 0", func() {
						resultMetrics, err := client.QueryMetrics(metricMappingConfig, emitLeaderFollowerMetrics)
						Expect(err).ToNot(HaveOccurred())

						metric := findMetricByKey(resultMetrics, "/p.mysql/follower/is_follower")
						Expect(metric).NotTo(BeNil())
						Expect(metric.Value).To(Equal(0.0))
						Expect(metric.Unit).To(Equal("bool"))
					})
				})

				Describe("when SHOW SLAVE STATUS returns a non-empty result", func() {
					BeforeEach(func() {
						emitLeaderFollowerMetrics = true
						seedDatabase()
					})

					Describe("the is_follower metric", func() {
						BeforeEach(func() {
							seedSlaveStatusTwice("some-value")
						})

						It("is emitted with value 1.0", func() {
							resultMetrics, err := client.QueryMetrics(metricMappingConfig, emitLeaderFollowerMetrics)
							Expect(err).ToNot(HaveOccurred())

							metric := findMetricByKey(resultMetrics, "/p.mysql/follower/is_follower")
							Expect(metric).NotTo(BeNil())
							Expect(metric.Value).To(Equal(1.0))
							Expect(metric.Unit).To(Equal("bool"))
						})
					})

					Describe("when seconds_behind_master is a number", func() {
						BeforeEach(func() {
							seedSlaveStatusTwice("2")
						})

						It("emits the seconds_behind_master metric with value equal to the number", func() {
							resultMetrics, err := client.QueryMetrics(metricMappingConfig, emitLeaderFollowerMetrics)
							Expect(err).ToNot(HaveOccurred())

							metric := findMetricByKey(resultMetrics, "/p.mysql/follower/seconds_behind_master")
							Expect(metric).NotTo(BeNil())
							Expect(metric.Value).To(Equal(2.0))
							Expect(metric.Unit).To(Equal("number"))
						})
					})

					Describe("when seconds_behind_master is NULL", func() {
						BeforeEach(func() {
							seedSlaveStatusTwice("NULL")
						})

						It("emits the seconds_behind_master metric with value equal to -1 (TBD)", func() {
							resultMetrics, err := client.QueryMetrics(metricMappingConfig, emitLeaderFollowerMetrics)
							Expect(err).ToNot(HaveOccurred())

							metric := findMetricByKey(resultMetrics, "/p.mysql/follower/seconds_behind_master")
							Expect(metric).NotTo(BeNil())
							Expect(metric.Value).To(Equal(-1.0))
							Expect(metric.Unit).To(Equal("number"))
						})
					})
				})
			})

			Describe("when emitLeaderFollowerMetrics is false", func() {
				BeforeEach(func() {
					emitLeaderFollowerMetrics = false
					seedDatabase()
				})

				It("should not emit is_follower metric", func() {
					resultMetrics, err := client.QueryMetrics(metricMappingConfig, emitLeaderFollowerMetrics)
					Expect(err).ToNot(HaveOccurred())
					for _, metric := range resultMetrics {
						Expect(metric.Key).ToNot(Equal("/p.mysql/follower/is_follower"))
					}
				})
			})

			Describe("when recovering from a prior bad parse", func() {
				BeforeEach(func() {
					emitLeaderFollowerMetrics = false
				})
				It("should treat the metric as brand new with no history of the past parse", func() {
					seedOneRowFromDatabase("wsrep_cluster_status", "This Breaks Parse")

					resultMetrics, err := client.QueryMetrics(metricMappingConfig, emitLeaderFollowerMetrics)
					Expect(err).NotTo(HaveOccurred())
					Expect(resultMetrics).NotTo(BeNil())

					metric := findMetricByKey(resultMetrics, "wsrep_cluster_status")
					Expect(metric.RawValue).To(Equal("This Breaks Parse"))
					Expect(metric.HasError).To(BeTrue())

					seedOneRowFromDatabase("wsrep_cluster_status", "Primary")

					resultMetrics, err = client.QueryMetrics(metricMappingConfig, emitLeaderFollowerMetrics)
					Expect(err).NotTo(HaveOccurred())
					Expect(resultMetrics).NotTo(BeNil())

					metric = findMetricByKey(resultMetrics, "wsrep_cluster_status")
					Expect(metric.HasError).To(BeFalse())
					Expect(metric.Value).To(Equal(1.0))
				})
			})

			Describe("when the data needs a special mapping", func() {
				BeforeEach(func() {
					seedDatabase()
				})

				It("should map cluster status to a number", func() {
					resultMetrics, err := client.QueryMetrics(metricMappingConfig, emitLeaderFollowerMetrics)

					Expect(err).NotTo(HaveOccurred())
					Expect(resultMetrics).NotTo(BeNil())

					metric := findMetricByKey(resultMetrics, "wsrep_cluster_status")
					Expect(metric).NotTo(BeNil())
					// Primary maps to 1
					Expect(metric.Value).To(Equal(1.0))
					Expect(metric.Unit).To(Equal("number"))

					seedOneRowFromDatabase("wsrep_cluster_status", "Non-Primary")

					resultMetrics, err = client.QueryMetrics(metricMappingConfig, emitLeaderFollowerMetrics)
					metric = findMetricByKey(resultMetrics, "wsrep_cluster_status")
					Expect(metric).NotTo(BeNil())
					// Non-Primary maps to 0
					Expect(metric.Value).To(Equal(0.0))

					seedOneRowFromDatabase("wsrep_cluster_status", "Disconnected")

					resultMetrics, err = client.QueryMetrics(metricMappingConfig, emitLeaderFollowerMetrics)
					metric = findMetricByKey(resultMetrics, "wsrep_cluster_status")
					Expect(metric).NotTo(BeNil())
					// Disconnected maps to -1
					Expect(metric.Value).To(Equal(-1.0))

					seedOneRowFromDatabase("wsrep_cluster_statuS", "DISCONNECTED")

					resultMetrics, err = client.QueryMetrics(metricMappingConfig, emitLeaderFollowerMetrics)
					metric = findMetricByKey(resultMetrics, "wsrep_cluster_status")
					Expect(metric).NotTo(BeNil())
					// Disconnected maps to -1
					Expect(metric.Value).To(Equal(-1.0))
					Expect(metric.HasError).To(BeFalse())
				})

				It("should map wsrep_ready to a 1 or 0", func() {
					resultMetrics, err := client.QueryMetrics(metricMappingConfig, emitLeaderFollowerMetrics)
					Expect(err).NotTo(HaveOccurred())

					metric := findMetricByKey(resultMetrics, "wsrep_ready")
					Expect(metric).NotTo(BeNil())
					// ON to 1
					Expect(metric.Value).To(Equal(1.0))
					Expect(metric.Unit).To(Equal("bool"))

					seedOneRowFromDatabase("wsrep_ready", "OFF")

					resultMetrics, err = client.QueryMetrics(metricMappingConfig, emitLeaderFollowerMetrics)
					metric = findMetricByKey(resultMetrics, "wsrep_ready")
					// OFF to 0
					Expect(metric.Value).To(Equal(0.0))

					seedOneRowFromDatabase("wsrep_ready", "FoobarValue")

					resultMetrics, err = client.QueryMetrics(metricMappingConfig, emitLeaderFollowerMetrics)
					metric = findMetricByKey(resultMetrics, "wsrep_ready")
					Expect(resultMetrics).NotTo(BeNil())
					metric = findMetricByKey(resultMetrics, "wsrep_ready")
					Expect(metric).NotTo(BeNil())
					Expect(metric.RawValue).To(Equal("FoobarValue"))
					Expect(metric.HasError).To(BeTrue())
				})
			})
		})
	})
})
